// VideoRecognition.cpp : detect and recognize faces in a video file.
//

#include "stdafx.h"
#include "resource.h"
#include "LuxandFaceSDK.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <windows.h>
#include <commctrl.h>

using namespace std;


extern "C" {
#include "libavutil/opt.h"
#include "libavcodec/avcodec.h"
#include "libavformat/avformat.h"
#include "libavutil/channel_layout.h"
#include "libavutil/common.h"
#include "libavutil/imgutils.h"
#include "libavutil/mathematics.h"
#include "libavutil/samplefmt.h"
#include "libavutil/imgutils.h"
#include "libavutil/timestamp.h"
#include "libswscale/swscale.h"
}

#pragma comment(lib, "avcodec.lib")
#pragma comment(lib, "avformat.lib")
#pragma comment(lib, "avutil.lib")


LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
INT_PTR CALLBACK	InputName(HWND, UINT, WPARAM, LPARAM);

const char * TrackerMemoryFile = "tracker70.dat";

enum ProgramState { psActive, psInactive, psRemember, psOpen, psTimechange };
ProgramState programState;
bool Pause, NeedClose;

HWND hwnd, TextBox, OpenButton, PauseButton, TimeSlider;
HDC dc, hbitmapDC;

typedef struct {
	HImage Picture;
	int64_t timestamp;
} Frame;

vector<Frame> Images;

long long IDs[256]; // detected faces
long long faceCount = 0;

RECT faceRects[256];
int MouseX = 0;
int MouseY = 0;

char userName[MAX_PATH];
char infilename[MAX_PATH];

int Width, Height;
int width = 640, height = 480;
double ratio;
double framerate;
int64_t duration;

int InitializeWindow() {
	LOGBRUSH brush;
	brush.lbColor = RGB(230, 230, 230);
	brush.lbStyle = BS_SOLID;
	HBRUSH bBrush;
	bBrush = CreateBrushIndirect(&brush);
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style = 0;
	wcex.lpfnWndProc = WndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = 0;
	wcex.hIcon = 0;
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = bBrush;
	wcex.lpszMenuName = 0;
	wcex.lpszClassName = L"My Window Class";
	wcex.hIconSm = 0;
	RegisterClassEx(&wcex);

	hwnd = CreateWindowEx(WS_EX_TOOLWINDOW, L"My Window Class", L"Video Face Recognition", WS_SYSMENU | WS_CAPTION | WS_SIZEBOX, 0, 0, 0, 0, 0, 0, 0, 0);
	dc = GetDC(hwnd);
	hbitmapDC = CreateCompatibleDC(dc);
	SetWindowPos(hwnd, 0, 100, 100, 6 + width, 6 + 16 + height + 60, SWP_NOZORDER);
	ShowWindow(hwnd, SW_SHOW);
	HPEN hNPen = CreatePen(PS_SOLID, 1, RGB(0, 255, 0));
	HPEN hOPen = (HPEN)SelectObject(dc, hNPen);
	HBRUSH hOldBrush;
	HBRUSH hNewBrush;
	hNewBrush = (HBRUSH)GetStockObject(NULL_BRUSH);
	hOldBrush = (HBRUSH)SelectObject(dc, hNewBrush);

	UpdateWindow(hwnd);

	HFONT MyFont = CreateFont(12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, L"Microsoft Sans Serif");

	TextBox = CreateWindow(L"Static", L"Click face to name it", SS_CENTER | WS_CHILD, 0, 0, 0, 0, hwnd, 0, 0, 0);
	SendMessage(TextBox, WM_SETFONT, WPARAM(MyFont), TRUE);
	SetWindowPos(TextBox, 0, 0, height + 20, width, 40, SWP_NOZORDER);
	ShowWindow(TextBox, SW_SHOW);
	UpdateWindow(hwnd);

	OpenButton = CreateWindow(L"BUTTON", L"Open Video File", WS_TABSTOP | WS_VISIBLE | WS_CHILD | BS_DEFPUSHBUTTON, 0, 0, 100, 100, hwnd, 0, 0, 0);
	SendMessage(OpenButton, WM_SETFONT, WPARAM(MyFont), TRUE);
	SetWindowPos(OpenButton, 0, width / 20, height + 25, width / 4, 30, SWP_NOZORDER);
	ShowWindow(OpenButton, SW_SHOW);
	UpdateWindow(hwnd);

	PauseButton = CreateWindow(L"BUTTON", L"Pause Video", WS_TABSTOP | WS_VISIBLE | WS_CHILD | BS_DEFPUSHBUTTON, 0, 0, 100, 100, hwnd, 0, 0, 0);
	SendMessage(PauseButton, WM_SETFONT, WPARAM(MyFont), TRUE);
	SetWindowPos(PauseButton, 0, width - width / 4 - width / 20 - 10, height + 25, width / 4, 30, SWP_NOZORDER);
	ShowWindow(PauseButton, SW_SHOW);
	UpdateWindow(hwnd);


	TimeSlider = CreateWindow(TRACKBAR_CLASS, 0, WS_CHILD | WS_VISIBLE | TBS_NOTICKS, 0, 0, 100, 100, hwnd, 0, 0, 0);
	SetWindowPos(TimeSlider, 0, 0, height, width, 20, SWP_NOZORDER);
	SendMessage(TimeSlider, TBM_SETRANGE, (WPARAM)TRUE, (LPARAM)MAKELONG(0, 1000));
	SendMessage(TimeSlider, TBM_SETSEL, (WPARAM)FALSE, (LPARAM)MAKELONG(0, 1000));
	SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)0);
	ShowWindow(TimeSlider, SW_SHOW);
	UpdateWindow(hwnd);
	return 0;
}

int GetSelectedRectangle() {
	int RectSelected = -1;
	for (int i = 0; i < faceCount; i++)
		if (MouseX >= faceRects[i].left && MouseX <= faceRects[i].right &&
			MouseY >= faceRects[i].top && MouseY <= faceRects[i].bottom) {
			RectSelected = i;
			break;
		}

	if (RectSelected >= 0) {
		HCURSOR Cur = LoadCursor(NULL, IDC_HAND);
		SetCursor(Cur);
	}
	return RectSelected;
}

void DisplayImage(HImage imageHandle) {
	HBITMAP hbitmapHandle; // to store the HBITMAP handle
	FSDK_SaveImageToHBitmap(imageHandle, &hbitmapHandle);
	int dx = width - Width * ratio + 1e-5;
	int dy = height - Height * ratio + 1e-5;
	HBRUSH brush = CreateSolidBrush(RGB(230, 230, 230));
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = width;
	rect.bottom = dy / 2;
	FillRect(dc, &rect, brush);
	rect.left = 0;
	rect.top = 0;
	rect.right = dx / 2;
	rect.bottom = height;
	FillRect(dc, &rect, brush);
	rect.left = 0;
	rect.top = height - dy + dy / 2;
	rect.right = width;
	rect.bottom = height;
	FillRect(dc, &rect, brush);
	rect.left = width - dx + dx / 2;
	rect.top = 0;
	rect.right = width;
	rect.bottom = height;
	FillRect(dc, &rect, brush);
	DeleteObject(brush);

	SelectObject(hbitmapDC, hbitmapHandle);
	SetStretchBltMode(dc, COLORONCOLOR);
	StretchBlt(dc, dx / 2, dy / 2, width - dx, height - dy, hbitmapDC, 0, 0, Width, Height, SRCCOPY);

	DeleteObject(hbitmapHandle); // delete the HBITMAP object
}

void DisplayText(char * name, RECT r) {
	SetBkMode(dc, TRANSPARENT);
	SetTextColor(dc, RGB(0, 255, 0));
	HFONT MyFont = CreateFont(25, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, L"Microsoft Sans Serif");
	SelectObject(dc, MyFont);
	DrawTextA(dc, name, -1, &r, DT_CENTER);
	DeleteObject(MyFont);
}

void ClearImage() {
	HBRUSH brush = CreateSolidBrush(RGB(230, 230, 230));
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = width;
	rect.bottom = height;
	FillRect(dc, &rect, brush);
	DeleteObject(brush);
}


static int SaveFrame(AVFrame* frame)
{
	if (frame->width != Width || frame->height != Height)
		return -1;

	struct SwsContext *sws_ctx = NULL;
	sws_ctx = sws_getContext(Width, Height, (AVPixelFormat)(frame->format), Width, Height, AV_PIX_FMT_RGB24, SWS_BILINEAR, NULL, NULL, NULL);
	int Scalline = 3 * Width;
	unsigned char* buf = new unsigned char[Scalline * Height];
	sws_scale(sws_ctx, frame->data, frame->linesize, 0, frame->height, &buf, &Scalline);
	Frame img;
	img.timestamp = frame->pts;
	FSDK_LoadImageFromBuffer(&(img.Picture), buf, Width, Height, 3 * Width, FSDK_IMAGE_COLOR_24BIT);
	Images.push_back(img);
	delete buf;
}

bool DecodePacket(AVCodecContext* dec, const AVPacket* pkt, AVFrame* frame) {
	int err = 0;

	// submit the packet to the decoder
	err = avcodec_send_packet(dec, pkt);
	if (err < 0)
		return FALSE;

	// get all the available frames from the decoder
	while (err >= 0) {
		err = avcodec_receive_frame(dec, frame);
		if (err < 0)
			if (err == AVERROR_EOF || err == AVERROR(EAGAIN))
				return TRUE;
			else
				return FALSE;


		// save the frame data to images vector
		err = SaveFrame(frame);

		av_frame_unref(frame);
		if (err < 0)
			return FALSE;
	}

	return TRUE;
}

bool OpenVideoCodecContext(int* stream_idx, AVCodecContext** dec_ctx, AVFormatContext* fmt_ctx, enum AVMediaType type) {
	int stream_index = av_find_best_stream(fmt_ctx, type, -1, -1, NULL, 0);
	if (stream_index < 0)
		return FALSE;

	AVStream* st = fmt_ctx->streams[stream_index];

	// find decoder for the stream 
	const AVCodec* dec = avcodec_find_decoder(st->codecpar->codec_id);
	if (!dec)
		return FALSE;

	*dec_ctx = avcodec_alloc_context3(dec);
	if (!*dec_ctx)
		return FALSE;

	// Copy codec parameters from input stream to output codec context and init the decoders
	if ((avcodec_parameters_to_context(*dec_ctx, st->codecpar) < 0) || (avcodec_open2(*dec_ctx, dec, NULL) < 0)) {
		avcodec_free_context(dec_ctx);
		return FALSE;
	}

	*stream_idx = stream_index;

	return TRUE;
}

inline void ClearData(AVFormatContext** fmt_ctx, AVCodecContext** video_dec_ctx) {
	avcodec_free_context(video_dec_ctx);
	avformat_close_input(fmt_ctx);
}

int _tmain(int argc, _TCHAR* argv[]) {

	if (FSDKE_OK != FSDK_ActivateLibrary("")) {
		MessageBox(0, L"Please insert a valid activation key\n", L"Error activating FaceSDK", MB_ICONERROR | MB_OK);
		exit(-1);
	}

	FSDK_Initialize("");

	// creating a Tracker
	HTracker tracker = 0;
	char CurDir[MAX_PATH];
	GetCurrentDirectoryA(sizeof(CurDir), CurDir);
	char TrackerMemoryPath[MAX_PATH];
	sprintf_s(TrackerMemoryPath, "%s\\%s", CurDir, TrackerMemoryFile);
	if (0 != FSDK_LoadTrackerMemoryFromFile(&tracker, TrackerMemoryPath)) // try to load saved tracker state
		FSDK_CreateTracker(&tracker); // if could not be loaded, create a new tracker

	int err = 0;
	FSDK_SetTrackerMultipleParameters(tracker, "HandleArbitraryRotations=false; DetermineFaceRotationAngle=false; InternalResizeWidth=100; FaceDetectionThreshold=5;", &err); // set realtime face detection parameters

	AVFrame* frame = av_frame_alloc();
	AVPacket* pkt = av_packet_alloc();
	AVFormatContext* fmt_ctx = NULL;
	AVCodecContext* video_dec_ctx = NULL;
	int video_stream_idx = -1;

	InitializeWindow();

	MSG msg = { 0 };
	NeedClose = FALSE;
	Pause = TRUE;
	programState = psInactive;
	bool VideoFinishes;

	vector<Frame>::iterator Img = Images.end();

	while (msg.message != WM_QUIT) {
		// process messages to find if the user clicked on a face
		if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			if ((msg.message == WM_KEYDOWN && msg.wParam == VK_ESCAPE) || NeedClose)
				break;
		}
		

		if (psOpen == programState) {
			ClearData(&fmt_ctx, &video_dec_ctx);
			printf("Opening Video File...\n");
			if ((avformat_open_input(&fmt_ctx, infilename, NULL, NULL) >= 0) && (avformat_find_stream_info(fmt_ctx, NULL) >= 0) && (OpenVideoCodecContext(&video_stream_idx, &video_dec_ctx, fmt_ctx, AVMEDIA_TYPE_VIDEO))) {
				Width = video_dec_ctx->width;
				Height = video_dec_ctx->height;
				ratio = min((double)width / Width, (double)height / Height);
				framerate = av_q2d(fmt_ctx->streams[video_stream_idx]->r_frame_rate);
				duration = fmt_ctx->streams[video_stream_idx]->duration;
				SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)0);

				Img = Images.end();
				programState = psActive;
				Pause = FALSE;
				SetWindowText(PauseButton, L"Pause Video");
				VideoFinishes = FALSE;
			}
			else {
				for (Img = Images.begin(); Img != Images.end(); ++Img)
					FSDK_FreeImage(Img->Picture);
				Images.clear();
				ClearImage();
				SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)0);
				programState = psInactive;
				MessageBox(0, L"Error occurs during opening and decoding video file! Try another one please.\n", L"Error Opening Video File", MB_ICONERROR | MB_OK);
			}
		}
		else if (psTimechange == programState)
		{
			for (Img = Images.begin(); Img != Images.end(); ++Img)
				FSDK_FreeImage(Img->Picture);
			Images.clear();
			Img = Images.end();
			VideoFinishes = FALSE;
			int64_t timestamp = SendMessage(TimeSlider, TBM_GETPOS, (WPARAM)0, (LPARAM)0) * duration / 1000;
			av_seek_frame(fmt_ctx, video_stream_idx, timestamp, AVSEEK_FLAG_ANY);
			programState = psActive;
		}
		else if ((!Pause) && (psActive == programState)) {
			DWORD starttime = GetTickCount();
			if (Img == Images.end()) {
				for (Img = Images.begin(); Img != Images.end(); ++Img)
					FSDK_FreeImage(Img->Picture);
				Images.clear();
				while ((!VideoFinishes) && (Images.size() < 1)) {
					int err = av_read_frame(fmt_ctx, pkt);
					if (err >= 0) {
						// check if the packet belongs to a video stream, otherwise skip it
						if (pkt->stream_index == video_stream_idx)
							err = DecodePacket(video_dec_ctx, pkt, frame);
						av_packet_unref(pkt);
					}
					else {
						if (video_dec_ctx)
							DecodePacket(video_dec_ctx, NULL, frame);
						VideoFinishes = TRUE;
					}
				}
				Img = Images.begin();
			}
			FSDK_FeedFrame(tracker, 0, Img->Picture, &faceCount, IDs, sizeof(IDs));
			DisplayImage(Img->Picture);
			if (Img->timestamp >= 0)
				SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)(1000 * Img->timestamp) / duration);

			for (int i = 0; i < faceCount; i++) {
				TFacePosition facePosition;
				FSDK_GetTrackerFacePosition(tracker, 0, IDs[i], &facePosition);

				facePosition.xc = ratio * facePosition.xc + (width - Width * ratio) / 2;
				facePosition.yc = ratio * facePosition.yc + (height - Height * ratio) / 2;
				facePosition.w = ratio * facePosition.w;

				faceRects[i].left = facePosition.xc - (int)(facePosition.w*0.6);
				faceRects[i].top = min(height, facePosition.yc - (int)(facePosition.w*0.5));
				faceRects[i].right = facePosition.xc + (int)(facePosition.w*0.6);
				faceRects[i].bottom = min(height, facePosition.yc + (int)(facePosition.w*0.7));
				Rectangle(dc, faceRects[i].left, faceRects[i].top, faceRects[i].right, faceRects[i].bottom);

				char name[1024];
				int res = FSDK_GetAllNames(tracker, IDs[i], name, sizeof(name));

				if (0 == res && strlen(name) > 0) {
					// draw name
					RECT NameRect;
					NameRect.left = faceRects[i].left - width;
					NameRect.right = faceRects[i].right + width;
					NameRect.top = min(height, faceRects[i].bottom);
					NameRect.bottom = min(height, faceRects[i].bottom + 30);
					DisplayText(name, NameRect);
				}
			}

			if (Img != Images.end())
				++Img;
			if ((Img == Images.end()) && (VideoFinishes))
			{
				ClearImage();
				SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)0);
				programState = psInactive;
			}

			DWORD worktime = GetTickCount() - starttime;
			if (worktime < 1000 / framerate)
				Sleep(1000 / framerate - worktime);
		}
		else
			Sleep(10);

		int RectSelected = GetSelectedRectangle(); // find out which face the mouse is over
		if (psRemember == programState) {
			if (RectSelected >= 0) {
				long long id = IDs[RectSelected];

				if (0 == FSDK_LockID(tracker, id)) {
					strcpy_s(userName, "_noname_");
					// get the user name
					DialogBox(NULL, MAKEINTRESOURCE(IDD_DIALOGINPUTNAME), NULL, InputName);
					if (strcmp(userName, "_noname_") != 0)
						FSDK_SetName(tracker, IDs[RectSelected], userName);
					if (strcmp(userName, "") == 0)
						FSDK_PurgeID(tracker, IDs[RectSelected]);
					FSDK_UnlockID(tracker, IDs[RectSelected]);
				}
			}
			programState = psActive;
		}
	}

	ReleaseDC(hwnd, dc);

	ClearData(&fmt_ctx, &video_dec_ctx);
	av_packet_free(&pkt);
	av_frame_free(&frame);

	FSDK_SaveTrackerMemoryToFile(tracker, TrackerMemoryPath);
	FSDK_FreeTracker(tracker);

	for (Img = Images.begin(); Img != Images.end(); ++Img)
		FSDK_FreeImage(Img->Picture);



	return 0;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_MOUSEMOVE:
		MouseX = LOWORD(lParam);
		MouseY = HIWORD(lParam);
		return (INT_PTR)TRUE;

	case WM_LBUTTONUP:
		if (programState == psActive)
			programState = psRemember;
		return (INT_PTR)TRUE;

	case WM_CTLCOLORSTATIC:
		SetBkMode((HDC)wParam, TRANSPARENT);
		return (int)CreateSolidBrush(RGB(230, 230, 230));

	case WM_SIZE:
		width = LOWORD(lParam);
		height = HIWORD(lParam) - 60;
		ratio = min((double)width / Width, (double)height / Height);
		SetWindowPos(TextBox, 0, 0, height + 20, width, 40, SWP_NOZORDER);
		ShowWindow(TextBox, SW_HIDE);
		ShowWindow(TextBox, SW_SHOWNORMAL);
		SetWindowPos(OpenButton, 0, width / 20, height + 25, width / 4, 30, SWP_NOZORDER);
		ShowWindow(OpenButton, SW_HIDE);
		ShowWindow(OpenButton, SW_SHOWNORMAL);
		SetWindowPos(PauseButton, 0, width - width / 4 - width / 20 - 10, height + 25, width / 4, 30, SWP_NOZORDER);
		ShowWindow(PauseButton, SW_HIDE);
		ShowWindow(PauseButton, SW_SHOWNORMAL);
		SetWindowPos(TimeSlider, 0, 0, height, width, 20, SWP_NOZORDER);
		ShowWindow(TimeSlider, SW_HIDE);
		ShowWindow(TimeSlider, SW_SHOWNORMAL);
		return (INT_PTR)TRUE;

	case WM_COMMAND:
		if (wParam == BN_CLICKED)
			if (lParam == (LPARAM)OpenButton) {
				OPENFILENAMEA oFile;
				char szPath[MAX_PATH];
				char szDir[MAX_PATH];
				ZeroMemory(&oFile, sizeof(oFile));
				GetCurrentDirectoryA(sizeof(szDir), szDir);
				lstrcpyA(szPath, "");
				oFile.lStructSize = sizeof(oFile);
				oFile.hwndOwner = hWnd;
				oFile.hInstance = NULL;
				oFile.lpstrFilter = "Video(*.*)\0*.*\0\0";
				oFile.lpstrFile = szPath;
				oFile.nMaxFile = sizeof(szPath);
				oFile.lpstrInitialDir = szDir;
				oFile.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
				oFile.lpstrDefExt = NULL;
				GetOpenFileNameA(&oFile);
				strcpy_s(infilename, oFile.lpstrFile);
				programState = psOpen;
			}
			else if (lParam == (LPARAM)PauseButton)
			{
				Pause = !Pause;
				if (Pause)
					SetWindowText(PauseButton, L"Play Video");
				else
					SetWindowText(PauseButton, L"Pause Video");
			}
		return (INT_PTR)TRUE;

	case WM_HSCROLL:
		if (lParam == (LPARAM)TimeSlider) {
			if (psActive == programState) {
				if ((SB_PAGERIGHT == LOWORD(wParam)) || (SB_PAGELEFT == LOWORD(wParam))) {
					POINT curPos;
					RECT SliderRect, wndSliderRect;
					GetCursorPos(&curPos);
					GetWindowRect(TimeSlider, &wndSliderRect);
					SendMessage(TimeSlider, TBM_GETCHANNELRECT, (WPARAM)0, (LPARAM)&SliderRect);
					int MaxRange = SendMessage(TimeSlider, TBM_GETRANGEMAX, (WPARAM)0, (LPARAM)0);
					int MinRange = SendMessage(TimeSlider, TBM_GETRANGEMIN, (WPARAM)0, (LPARAM)0);
					int SliderLength = SliderRect.right - SliderRect.left;
					int TotalRange = MaxRange - MinRange;
					int curX = curPos.x - wndSliderRect.left - SliderRect.left;
					int pos = MinRange + (TotalRange * curX) / SliderLength;
					SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)pos);
				}
				programState = psTimechange;
			}
			else if (psInactive == programState)
				SendMessage(TimeSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)0);
		}
		return (INT_PTR)TRUE;


	case WM_DESTROY:
		NeedClose = TRUE;
		break;
	}
	return DefWindowProc(hWnd, message, wParam, lParam);
}

INT_PTR CALLBACK InputName(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_SHOWWINDOW:
		SetFocus(GetDlgItem(hDlg, IDC_EDIT1));
		return (INT_PTR)TRUE;
		break;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK) {
			int len = SendDlgItemMessageA(hDlg, IDC_EDIT1, WM_GETTEXT, (WPARAM) sizeof(userName) / 2 - 1, (LPARAM)userName);
			userName[len] = '\0';
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;
		}
		if (LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg, LOWORD(wParam));
			return (INT_PTR)TRUE;
		}
		break;
	}
	return (INT_PTR)FALSE;
}
